@@ -1,29 +1,57 @@
 # -*- coding: utf-8 -*-
 '''
 Support for iptables
+
+Configuration Options
+---------------------
+
+The following options can be set in the minion config, grains, pillar, or
+master config. The configuration is read using :py:func:`config.get
+<salt.modules.config.get>`.
+
+- ``iptables.save_filters``: List of REGEX strings to FILTER OUT matching lines
+
+  This is useful for filtering out chains, rules, etc that you do not wish to
+  persist, such as ephemeral Docker rules.
+
+  The default is to not filter out anything.
+
+  .. code-block:: yaml
+
+      iptables.save_filters:
+        - "-j CATTLE_PREROUTING"
+        - "-j DOCKER"
+        - "-A POSTROUTING"
+        - "-A CATTLE_POSTROUTING"
+        - "-A FORWARD"
 '''
-from __future__ import absolute_import
+from __future__ import absolute_import, unicode_literals, print_function
 
 # Import python libs
 import os
 import re
 import sys
 import uuid
-import shlex
 import string
 
 # Import salt libs
-import salt.utils
+import salt.utils.args
+import salt.utils.files
+import salt.utils.path
 from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS
 from salt.exceptions import SaltException
+from salt.ext import six
+
+import logging
+log = logging.getLogger(__name__)
 
 
 def __virtual__():
     '''
     Only load the module if iptables is installed
     '''
-    if not salt.utils.which('iptables'):
-        return False
+    if not salt.utils.path.which('iptables'):
+        return (False, 'The iptables execution module cannot be loaded: iptables not installed.')
 
     return True
 
@@ -33,9 +61,9 @@ def _iptables_cmd(family='ipv4'):
     Return correct command based on the family, e.g. ipv4 or ipv6
     '''
     if family == 'ipv6':
-        return salt.utils.which('ip6tables')
+        return salt.utils.path.which('ip6tables')
     else:
-        return salt.utils.which('iptables')
+        return salt.utils.path.which('iptables')
 
 
 def _has_option(option, family='ipv4'):
@@ -80,12 +108,72 @@ def _conf(family='ipv4'):
     elif __grains__['os_family'] == 'Suse':
         # SuSE does not seem to use separate files for IPv4 and IPv6
         return '/etc/sysconfig/scripts/SuSEfirewall2-custom'
+    elif __grains__['os_family'] == 'Void':
+        if family == 'ipv4':
+            return '/etc/iptables/iptables.rules'
+        else:
+            return '/etc/iptables/ip6tables.rules'
+    elif __grains__['os'] == 'Alpine':
+        if family == 'ipv6':
+            return '/etc/iptables/rules6-save'
+        else:
+            return '/etc/iptables/rules-save'
     else:
         raise SaltException('Saving iptables to file is not' +
                             ' supported on {0}.'.format(__grains__['os']) +
                             ' Please file an issue with SaltStack')
 
 
+def _conf_save_filters():
+    '''
+    Return array of strings from `save_filters` in config.
+
+    This array will be pulled from minion config, minion grains,
+    minion pillar, or master config.  The default value returned is [].
+
+    .. code-block:: python
+
+        _conf_save_filters()
+    '''
+    config = __salt__['config.option']('iptables.save_filters', [])
+    return config
+
+
+def _regex_iptables_save(cmd_output, filters=None):
+    '''
+    Return string with `save_filter` regex entries removed.  For example:
+
+    If `filters` is not provided, it will be pulled from minion config,
+    minion grains, minion pillar, or master config. Default return value
+    if no filters found is the original cmd_output string.
+
+    .. code-block:: python
+
+        _regex_iptables_save(cmd_output, ['-A DOCKER*'])
+    '''
+    # grab RE compiled filters from context for performance
+    if 'iptables.save_filters' not in __context__:
+        __context__['iptables.save_filters'] = []
+        for pattern in filters or _conf_save_filters():
+            try:
+                __context__['iptables.save_filters']\
+                    .append(re.compile(pattern))
+            except re.error as e:
+                log.warning('Skipping regex rule: \'%s\': %s',
+                            pattern, e)
+                continue
+
+    if len(__context__['iptables.save_filters']) > 0:
+        # line by line get rid of any regex matches
+        _filtered_cmd_output = \
+            [line for line in cmd_output.splitlines(True)
+             if not any(reg.search(line) for reg
+                        in __context__['iptables.save_filters'])]
+        return ''.join(_filtered_cmd_output)
+
+    return cmd_output
+
+
 def version(family='ipv4'):
     '''
     Return version from iptables --version
@@ -104,7 +192,7 @@ def version(family='ipv4'):
     return out[1]
 
 
-def build_rule(table=None, chain=None, command=None, position='', full=None, family='ipv4',
+def build_rule(table='filter', chain=None, command=None, position='', full=None, family='ipv4',
                **kwargs):
     '''
     Build a well-formatted iptables rule based on kwargs. A `table` and `chain`
@@ -118,11 +206,20 @@ def build_rule(table=None, chain=None, command=None, position='', full=None, fam
     If a position is required (as with `-I` or `-D`), it may be specified as
     `position`. This will only be useful if `full` is True.
 
+    If `state` is passed, it will be ignored, use `connstate`.
     If `connstate` is passed in, it will automatically be changed to `state`.
 
     To pass in jump options that doesn't take arguments, pass in an empty
     string.
 
+    .. note::
+
+        Whereas iptables will accept ``-p``, ``--proto[c[o[l]]]`` as synonyms
+        of ``--protocol``, if ``--proto`` appears in an iptables command after
+        the appearance of ``-m policy``, it is interpreted as the ``--proto``
+        option of the policy extension (see the iptables-extensions(8) man
+        page).
+
     CLI Examples:
 
     .. code-block:: bash
@@ -131,19 +228,19 @@ def build_rule(table=None, chain=None, command=None, position='', full=None, fam
             connstate=RELATED,ESTABLISHED jump=ACCEPT
 
         salt '*' iptables.build_rule filter INPUT command=I position=3 \\
-            full=True match=state state=RELATED,ESTABLISHED jump=ACCEPT
+            full=True match=state connstate=RELATED,ESTABLISHED jump=ACCEPT
 
         salt '*' iptables.build_rule filter INPUT command=A \\
-            full=True match=state state=RELATED,ESTABLISHED \\
+            full=True match=state connstate=RELATED,ESTABLISHED \\
             source='127.0.0.1' jump=ACCEPT
 
         .. Invert Rules
         salt '*' iptables.build_rule filter INPUT command=A \\
-            full=True match=state state=RELATED,ESTABLISHED \\
-            source='! 127.0.0.1' jump=ACCEPT
+            full=True match=state connstate=RELATED,ESTABLISHED \\
+            source='!127.0.0.1' jump=ACCEPT
 
         salt '*' iptables.build_rule filter INPUT command=A \\
-            full=True match=state state=RELATED,ESTABLISHED \\
+            full=True match=state connstate=RELATED,ESTABLISHED \\
             destination='not 127.0.0.1' jump=ACCEPT
 
         IPv6:
@@ -151,252 +248,283 @@ def build_rule(table=None, chain=None, command=None, position='', full=None, fam
             connstate=RELATED,ESTABLISHED jump=ACCEPT \\
             family=ipv6
         salt '*' iptables.build_rule filter INPUT command=I position=3 \\
-            full=True match=state state=RELATED,ESTABLISHED jump=ACCEPT \\
+            full=True match=state connstate=RELATED,ESTABLISHED jump=ACCEPT \\
             family=ipv6
-
     '''
     if 'target' in kwargs:
-        kwargs['jump'] = kwargs['target']
-        del kwargs['target']
+        kwargs['jump'] = kwargs.pop('target')
+
+    # Ignore name and state for this function
+    kwargs.pop('name', None)
+    kwargs.pop('state', None)
 
     for ignore in list(_STATE_INTERNAL_KEYWORDS) + ['chain', 'save', 'table']:
         if ignore in kwargs:
             del kwargs[ignore]
 
-    rule = ''
-    proto = False
-    bang_not_pat = re.compile(r'[!|not]\s?')
+    rule = []
+    protocol = False
+    bang_not_pat = re.compile(r'(!|not)\s?')
+
+    def maybe_add_negation(arg):
+        '''
+        Will check if the defined argument is intended to be negated,
+        (i.e. prefixed with '!' or 'not'), and add a '! ' to the rule.
+
+        The prefix will be removed from the value in the kwargs dict.
+        '''
+        value = str(kwargs[arg])
+        if value.startswith('!') or value.startswith('not'):
+            kwargs[arg] = re.sub(bang_not_pat, '', value)
+            return '! '
+        return ''
 
     if 'if' in kwargs:
-        if kwargs['if'].startswith('!') or kwargs['if'].startswith('not'):
-            kwargs['if'] = re.sub(bang_not_pat, '', kwargs['if'])
-            rule += '! '
-
-        rule += '-i {0} '.format(kwargs['if'])
+        rule.append('{0}-i {1}'.format(maybe_add_negation('if'), kwargs['if']))
         del kwargs['if']
 
     if 'of' in kwargs:
-        if kwargs['of'].startswith('!') or kwargs['of'].startswith('not'):
-            kwargs['of'] = re.sub(bang_not_pat, '', kwargs['of'])
-            rule += '! '
-
-        rule += '-o {0} '.format(kwargs['of'])
+        rule.append('{0}-o {1}'.format(maybe_add_negation('of'), kwargs['of']))
         del kwargs['of']
 
-    if 'protocol' in kwargs:
-        proto = kwargs['protocol']
-        del kwargs['protocol']
-    elif 'proto' in kwargs:
-        proto = kwargs['proto']
+    if 'proto' in kwargs and kwargs.get('match') != 'policy':
+        kwargs['protocol'] = kwargs['proto']
         del kwargs['proto']
-
-    if proto:
-        if proto.startswith('!') or proto.startswith('not'):
-            proto = re.sub(bang_not_pat, '', proto)
-            rule += '! '
-
-        rule += '-p {0} '.format(proto)
-        proto = True
+        # Handle the case 'proto' in kwargs and kwargs.get('match') == 'policy' below
+    if 'protocol' in kwargs:
+        if not protocol:
+            rule.append('{0}-p {1}'.format(maybe_add_negation('protocol'), kwargs['protocol']))
+            protocol = True
+        del kwargs['protocol']
 
     if 'match' in kwargs:
-        if not isinstance(kwargs['match'], list):
-            kwargs['match'] = kwargs['match'].split(',')
-        for match in kwargs['match']:
-            rule += '-m {0} '.format(match)
-            if 'name' in kwargs and match.strip() in ('pknock', 'quota2', 'recent'):
-                rule += '--name {0} '.format(kwargs['name'])
+        match_value = kwargs['match']
+        if not isinstance(match_value, list):
+            match_value = match_value.split(',')
+        for match in match_value:
+            rule.append('-m {0}'.format(match))
+            if 'name_' in kwargs and match.strip() in ('pknock', 'quota2', 'recent'):
+                rule.append('--name {0}'.format(kwargs['name_']))
+                del kwargs['name_']
+        if 'proto' in kwargs and kwargs.get('match') == 'policy':
+            rule.append('{0}--proto {1}'.format(maybe_add_negation('proto'), kwargs['proto']))
+            del kwargs['proto']
         del kwargs['match']
 
-    if 'name' in kwargs:
-        del kwargs['name']
-
-    if 'state' in kwargs:
-        del kwargs['state']
+    if 'match-set' in kwargs:
+        if isinstance(kwargs['match-set'], six.string_types):
+            kwargs['match-set'] = [kwargs['match-set']]
+        for match_set in kwargs['match-set']:
+            negative_match_set = ''
+            if match_set.startswith('!') or match_set.startswith('not'):
+                negative_match_set = '! '
+                match_set = re.sub(bang_not_pat, '', match_set)
+            rule.append('-m set {0}--match-set {1}'.format(negative_match_set, match_set))
+        del kwargs['match-set']
 
     if 'connstate' in kwargs:
         if '-m state' not in rule:
-            rule += '-m state '
+            rule.append('-m state')
 
-        if kwargs['connstate'].startswith('!') or kwargs['connstate'].startswith('not'):
-            kwargs['connstate'] = re.sub(bang_not_pat, '', kwargs['connstate'])
-            rule += '! '
+        rule.append('{0}--state {1}'.format(maybe_add_negation('connstate'), kwargs['connstate']))
 
-        rule += '--state {0} '.format(kwargs['connstate'])
         del kwargs['connstate']
 
     if 'dport' in kwargs:
-        if str(kwargs['dport']).startswith('!') or str(kwargs['dport']).startswith('not'):
-            kwargs['dport'] = re.sub(bang_not_pat, '', kwargs['dport'])
-            rule += '! '
-
-        rule += '--dport {0} '.format(kwargs['dport'])
+        rule.append('{0}--dport {1}'.format(maybe_add_negation('dport'), kwargs['dport']))
         del kwargs['dport']
 
     if 'sport' in kwargs:
-        if str(kwargs['sport']).startswith('!') or str(kwargs['sport']).startswith('not'):
-            kwargs['sport'] = re.sub(bang_not_pat, '', kwargs['sport'])
-            rule += '! '
-
-        rule += '--sport {0} '.format(kwargs['sport'])
+        rule.append('{0}--sport {1}'.format(maybe_add_negation('sport'), kwargs['sport']))
         del kwargs['sport']
 
-    if 'dports' in kwargs:
-        if '-m multiport' not in rule:
-            rule += '-m multiport '
-            if not proto:
-                return 'Error: proto must be specified'
+    for multiport_arg in ('dports', 'sports'):
+        if multiport_arg in kwargs:
+            if '-m multiport' not in rule:
+                rule.append('-m multiport')
+                if not protocol:
+                    return 'Error: protocol must be specified'
 
-        if isinstance(kwargs['dports'], list):
-            if [item for item in kwargs['dports'] if str(item).startswith('!') or str(item).startswith('not')]:
-                kwargs['dports'] = [re.sub(bang_not_pat, '', str(item)) for item in kwargs['dports']]
-                rule += '! '
-            dports = ','.join([str(i) for i in kwargs['dports']])
-        else:
-            if str(kwargs['dports']).startswith('!') or str(kwargs['dports']).startswith('not'):
-                dports = re.sub(bang_not_pat, '', kwargs['dports'])
-                rule += '! '
+            mp_value = kwargs[multiport_arg]
+            if isinstance(mp_value, list):
+                if any(i for i in mp_value if str(i).startswith('!') or str(i).startswith('not')):
+                    mp_value = [re.sub(bang_not_pat, '', str(item)) for item in mp_value]
+                    rule.append('!')
+                dports = ','.join(str(i) for i in mp_value)
             else:
-                dports = kwargs['dports']
+                if str(mp_value).startswith('!') or str(mp_value).startswith('not'):
+                    dports = re.sub(bang_not_pat, '', mp_value)
+                    rule.append('!')
+                else:
+                    dports = mp_value
 
-        rule += '--dports {0} '.format(dports)
-        del kwargs['dports']
-
-    if 'sports' in kwargs:
-        if '-m multiport' not in rule:
-            rule += '-m multiport '
-            if not proto:
-                return 'Error: proto must be specified'
-
-        if isinstance(kwargs['sports'], list):
-            if [item for item in kwargs['sports'] if str(item).startswith('!') or str(item).startswith('not')]:
-                kwargs['sports'] = [re.sub(bang_not_pat, '', str(item)) for item in kwargs['sports']]
-                rule += '! '
-            sports = ','.join([str(i) for i in kwargs['sports']])
-        else:
-            if str(kwargs['sports']).startswith('!') or str(kwargs['sports']).startswith('not'):
-                sports = re.sub(bang_not_pat, '', kwargs['sports'])
-                rule += '! '
-            else:
-                sports = kwargs['sports']
-
-        rule += '--sports {0} '.format(sports)
-        del kwargs['sports']
+            rule.append('--{0} {1}'.format(multiport_arg, dports))
+            del kwargs[multiport_arg]
 
     if 'comment' in kwargs:
-        rule += '--comment "{0}" '.format(kwargs['comment'])
+        if '-m comment' not in rule:
+            rule.append('-m comment')
+
+        rule.append('--comment "{0}"'.format(kwargs['comment']))
         del kwargs['comment']
 
     # --set in ipset is deprecated, works but returns error.
     # rewrite to --match-set if not empty, otherwise treat as recent option
     if 'set' in kwargs and kwargs['set']:
-        if kwargs['set'].startswith('!') or kwargs['set'].startswith('not'):
-            kwargs['set'] = re.sub(bang_not_pat, '', kwargs['set'])
-            rule += '! '
-
-        rule += '--match-set {0} '.format(kwargs['set'])
+        rule.append('{0}--match-set {1}'.format(maybe_add_negation('set'), kwargs['set']))
         del kwargs['set']
 
     # Jumps should appear last, except for any arguments that are passed to
     # jumps, which of course need to follow.
     after_jump = []
-    # List of options fetched from http://www.iptables.info/en/iptables-targets-and-jumps.html
+    # All jump arguments as extracted from man iptables-extensions, man iptables,
+    # man xtables-addons and http://www.iptables.info/en/iptables-targets-and-jumps.html
     after_jump_arguments = (
         'j',  # j and jump needs to be first
         'jump',
+
+        # IPTABLES
+        'add-set',
+        'and-mark',
+        'and-tos',
+        'checksum-fill',
         'clamp-mss-to-pmtu',
-        'ecn-tcp-remove',  # no arg
-        'mask',  # only used with either save-mark or restore-mark
-        'nodst',
-        'queue-num',
-        'reject-with',
-        'restore',  # no arg
-        'restore-mark',  # no arg
-        #'save',  # no arg, problematic name: How do we avoid collision with this?
-        'save-mark',  # no arg
-        'selctx',
-        'set-dscp',
-        'set-dscp-class',
-        'set-mss',
-        'set-tos',
-        'ttl-dec',
-        'ttl-inc',
-        'ttl-set',
-        'ulog-cprange',
-        'ulog-nlgroup',
-        'ulog-prefix',
-        'ulog-qthreshold',
         'clustermac',
-        'hash-init,'
+        'ctevents',
+        'ctmask',
+        'del-set',
+        'ecn-tcp-remove',
+        'exist',
+        'expevents',
+        'gateway',
+        'hash-init',
         'hashmode',
+        'helper',
+        'label',
         'local-node',
         'log-ip-options',
         'log-level',
         'log-prefix',
         'log-tcp-options',
         'log-tcp-sequence',
-        'new',  # no arg
+        'log-uid',
+        'mask',
+        'new',
+        'nfmask',
+        'nflog-group',
+        'nflog-prefix',
+        'nflog-range',
+        'nflog-threshold',
+        'nodst',
+        'notrack',
+        'on-ip',
+        'on-port',
+        'or-mark',
+        'or-tos',
+        'persistent',
+        'queue-balance',
+        'queue-bypass',
+        'queue-num',
+        'random',
+        'rateest-ewmalog',
+        'rateest-interval',
+        'rateest-name',
         'reject-with',
+        'restore',
+        'restore-mark',
+        #'save',  # no arg, problematic name: How do we avoid collision with this?
+        'save-mark',
+        'selctx',
         'set-class',
+        'set-dscp',
+        'set-dscp-class',
         'set-mark',
+        'set-mss',
+        'set-tos',
         'set-xmark',
+        'strip-options',
+        'timeout',
         'to',
         'to-destination',
-        'to-port',
         'to-ports',
         'to-source',
-        'total-nodes,'
         'total-nodes',
+        'tproxy-mark',
+        'ttl-dec',
+        'ttl-inc',
+        'ttl-set',
+        'type',
+        'ulog-cprange',
+        'ulog-nlgroup',
+        'ulog-prefix',
+        'ulog-qthreshold',
+        'xor-mark',
+        'xor-tos',
+        'zone',
+
+        # IPTABLES-EXTENSIONS
+        'dst-pfx',
+        'hl-dec',
+        'hl-inc',
+        'hl-set',
+        'hmark-dport-mask',
+        'hmark-dst-prefix',
+        'hmark-mod',
+        'hmark-offset',
+        'hmark-proto-mask',
+        'hmark-rnd',
+        'hmark-spi-mask',
+        'hmark-sport-mask',
+        'hmark-src-prefix',
+        'hmark-tuple',
+        'led-always-blink',
+        'led-delay',
+        'led-trigger-id',
+        'queue-cpu-fanout',
+        'src-pfx',
+
+        # WEB
+        'to-port',
+
+        # XTABLES
+        'addr',
+        'and-mask',
+        'delude',
+        'honeypot',
+        'or-mask',
+        'prefix',
+        'reset',
+        'reuse',
+        'set-mac',
+        'shift',
+        'static',
+        'tarpit',
+        'tname',
+        'ttl',
     )
     for after_jump_argument in after_jump_arguments:
         if after_jump_argument in kwargs:
             value = kwargs[after_jump_argument]
-            if any(ws_char in str(value) for ws_char in string.whitespace):
+            if value in (None, ''):  # options without arguments
+                after_jump.append('--{0}'.format(after_jump_argument))
+            elif any(ws_char in str(value) for ws_char in string.whitespace):
                 after_jump.append('--{0} "{1}"'.format(after_jump_argument, value))
             else:
                 after_jump.append('--{0} {1}'.format(after_jump_argument, value))
             del kwargs[after_jump_argument]
 
-    if 'log' in kwargs:
-        after_jump.append('--log {0} '.format(kwargs['log']))
-        del kwargs['log']
+    for key in kwargs:
+        negation = maybe_add_negation(key)
+        # don't use .items() since maybe_add_negation removes the prefix from
+        # the value in the kwargs, thus we need to fetch it after that has run
+        value = kwargs[key]
+        flag = '-' if len(key) == 1 else '--'
+        value = '' if value in (None, '') else ' {0}'.format(value)
+        rule.append('{0}{1}{2}{3}'.format(negation, flag, key, value))
 
-    if 'log-level' in kwargs:
-        after_jump.append('--log-level {0} '.format(kwargs['log-level']))
-        del kwargs['log-level']
+    rule += after_jump
 
-    if 'log-prefix' in kwargs:
-        after_jump.append('--log-prefix {0} '.format(kwargs['log-prefix']))
-        del kwargs['log-prefix']
-
-    if 'log-tcp-sequence' in kwargs:
-        after_jump.append('--log-tcp-sequence {0} '.format(kwargs['log-tcp-sequence']))
-        del kwargs['log-tcp-sequence']
-
-    if 'log-tcp-options' in kwargs:
-        after_jump.append('--log-tcp-options {0} '.format(kwargs['log-tcp-options']))
-        del kwargs['log-tcp-options']
-
-    if 'log-ip-options' in kwargs:
-        after_jump.append('--log-ip-options {0} '.format(kwargs['log-ip-options']))
-        del kwargs['log-ip-options']
-
-    if 'log-uid' in kwargs:
-        after_jump.append('--log-uid {0} '.format(kwargs['log-uid']))
-        del kwargs['log-uid']
-
-    for item in kwargs:
-        if str(kwargs[item]).startswith('!') or str(kwargs[item]).startswith('not'):
-            kwargs[item] = re.sub(bang_not_pat, '', kwargs[item])
-            rule += '! '
-
-        if len(item) == 1:
-            rule += '-{0} {1} '.format(item, kwargs[item])
-        else:
-            rule += '--{0} {1} '.format(item, kwargs[item])
-
-    rule += ' '.join(after_jump)
-
-    if full in ['True', 'true']:
+    if full:
         if not table:
             return 'Error: Table needs to be specified'
         if not chain:
@@ -412,9 +540,9 @@ def build_rule(table=None, chain=None, command=None, position='', full=None, fam
         wait = '--wait' if _has_option('--wait', family) else ''
 
         return '{0} {1} -t {2} {3}{4} {5} {6} {7}'.format(_iptables_cmd(family),
-               wait, table, flag, command, chain, position, rule)
+               wait, table, flag, command, chain, position, ' '.join(rule))
 
-    return rule
+    return ' '.join(rule)
 
 
 def get_saved_rules(conf_file=None, family='ipv4'):
@@ -542,11 +670,18 @@ def save(filename=None, family='ipv4'):
     if _conf() and not filename:
         filename = _conf(family)
 
+    log.debug('Saving rules to %s', filename)
+
     parent_dir = os.path.dirname(filename)
     if not os.path.isdir(parent_dir):
         os.makedirs(parent_dir)
     cmd = '{0}-save'.format(_iptables_cmd(family))
     ipt = __salt__['cmd.run'](cmd)
+
+    # regex out the output if configured with filters
+    if len(_conf_save_filters()) > 0:
+        ipt = _regex_iptables_save(ipt)
+
     out = __salt__['file.write'](filename, ipt)
     return out
 
@@ -580,7 +715,7 @@ def check(table='filter', chain=None, rule=None, family='ipv4'):
 
     if _has_option('--check', family):
         cmd = '{0} -t {1} -C {2} {3}'.format(ipt_cmd, table, chain, rule)
-        out = __salt__['cmd.run'](cmd)
+        out = __salt__['cmd.run'](cmd, output_loglevel='quiet')
     else:
         _chain_name = hex(uuid.getnode())
 
@@ -721,6 +856,9 @@ def append(table='filter', chain=None, rule=None, family='ipv4'):
         return 'Error: Rule needs to be specified'
 
     wait = '--wait' if _has_option('--wait', family) else ''
+    returnCheck = check(table, chain, rule, family)
+    if isinstance(returnCheck, bool) and returnCheck:
+        return False
     cmd = '{0} {1} -t {2} -A {3} {4}'.format(
             _iptables_cmd(family), wait, table, chain, rule)
     out = __salt__['cmd.run'](cmd)
@@ -764,13 +902,16 @@ def insert(table='filter', chain=None, position=None, rule=None, family='ipv4'):
         return 'Error: Rule needs to be specified'
 
     if position < 0:
-        rules = get_rules(family='ipv4')
+        rules = get_rules(family=family)
         size = len(rules[table][chain]['rules'])
         position = (size + position) + 1
         if position is 0:
             position = 1
 
     wait = '--wait' if _has_option('--wait', family) else ''
+    returnCheck = check(table, chain, rule, family)
+    if isinstance(returnCheck, bool) and returnCheck:
+        return False
     cmd = '{0} {1} -t {2} -I {3} {4} {5}'.format(
             _iptables_cmd(family), wait, table, chain, position, rule)
     out = __salt__['cmd.run'](cmd)
@@ -846,7 +987,7 @@ def _parse_conf(conf_file=None, in_mem=False, family='ipv4'):
 
     rules = ''
     if conf_file:
-        with salt.utils.fopen(conf_file, 'r') as ifile:
+        with salt.utils.files.fopen(conf_file, 'r') as ifile:
             rules = ifile.read()
     elif in_mem:
         cmd = '{0}-save' . format(_iptables_cmd(family))
@@ -858,6 +999,7 @@ def _parse_conf(conf_file=None, in_mem=False, family='ipv4'):
     table = ''
     parser = _parser()
     for line in rules.splitlines():
+        line = salt.utils.stringutils.to_unicode(line)
         if line.startswith('*'):
             table = line.replace('*', '')
             ret[table] = {}
@@ -873,7 +1015,7 @@ def _parse_conf(conf_file=None, in_mem=False, family='ipv4'):
             ret[table][chain]['rules'] = []
             ret[table][chain]['rules_comment'] = {}
         elif line.startswith('-A'):
-            args = shlex.split(line)
+            args = salt.utils.args.shlex_split(line)
             index = 0
             while index + 1 < len(args):
                 swap = args[index] == '!' and args[index + 1].startswith('-')
@@ -893,11 +1035,8 @@ def _parse_conf(conf_file=None, in_mem=False, family='ipv4'):
             if args[-1].startswith('-'):
                 args.append('')
             parsed_args = []
-            if sys.version.startswith('2.6'):
-                (opts, leftover_args) = parser.parse_args(args)
-                parsed_args = vars(opts)
-            else:
-                parsed_args = vars(parser.parse_args(args))
+            opts, _ = parser.parse_known_args(args)
+            parsed_args = vars(opts)
             ret_args = {}
             chain = parsed_args['append']
             for arg in parsed_args:
@@ -912,9 +1051,9 @@ def _parse_conf(conf_file=None, in_mem=False, family='ipv4'):
 
 def _parser():
     '''
-    This function contains _all_ the options I could find in man 8 iptables,
-    listed in the first section that I found them in. They will not all be used
-    by all parts of the module; use them intelligently and appropriately.
+    This function attempts to list all the options documented in the
+    iptables(8) and iptables-extensions(8) man pages.  They will not all be
+    used by all parts of the module; use them intelligently and appropriately.
     '''
     add_arg = None
     if sys.version.startswith('2.6'):
@@ -963,6 +1102,8 @@ def _parser():
     add_arg('--ahres', dest='ahres', action='append')
     ## bpf
     add_arg('--bytecode', dest='bytecode', action='append')
+    ## cgroup
+    add_arg('--cgroup', dest='cgroup', action='append')
     ## cluster
     add_arg('--cluster-total-nodes',
             dest='cluster-total-nodes',
@@ -1334,6 +1475,8 @@ def _parser():
     add_arg('--or-mark', dest='or-mark', action='append')
     add_arg('--xor-mark', dest='xor-mark', action='append')
     add_arg('--set-mark', dest='set-mark', action='append')
+    add_arg('--nfmask', dest='nfmask', action='append')
+    add_arg('--ctmask', dest='ctmask', action='append')
     ## CONNSECMARK
     add_arg('--save', dest='save', action='append')
     add_arg('--restore', dest='restore', action='append')
